import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import os
from pathlib import Path
import io
from skimage.io import imread, imsave
import glob
from factory import *
from skimage.filters import gaussian
from skimage.transform import rescale, rotate
from utils import *
import argparse
from joblib import dump

parser = argparse.ArgumentParser()
parser.add_argument("-nt", type=int)
parser.add_argument("-nd", type=int)
parser.add_argument("-nc", type=int)
parser.add_argument("-name")
args = vars(parser.parse_args())

path = Path.cwd()
datapath = str(path) + '/'
print(datapath)

# coin parameters
size = 1024
coin_radius = size // 2
die_radius = size // 2 - 50

# load all available symbols
small = []
for symbol in glob.glob(datapath + 'symbols/small/*.png'):
  S = imread(symbol, 1)
  small.append(S)
small = np.array(small)
large = []
for symbol in glob.glob(datapath + 'symbols/large/*.png'):
  S = imread(symbol, 1)
  large.append(S)

# load all textures
textures = []
for tex in glob.glob(datapath + 'textures/*.png'):
  texture = imread(tex, 1)
  texture = normalize(texture.astype(float))
  textures.append(texture)
# tex_coeffs = [0.1, 0.05, 0.008, 0.01]
tex_coeffs = [0.05, 0.02, 0.0075, 0.02]

# conception parameters
# n_types = args["nt"]
# n_dies_per_type = args["nd"]
# n_coins_per_die = args["nc"]
n_types = 16
n_dies_per_type = 1
n_coins_per_die = 5
name = "batch_for_center_detection"
total_coins = n_types * n_dies_per_type * n_coins_per_die

large_index = np.random.permutation(len(large))

coin_database = {nt:
                 {nd:
                  {nc: 0 for nc in range(n_coins_per_die)}
                  for nd in range(n_dies_per_type)}
                 for nt in range(n_types)}

for nt in range(n_types):
  # choose symbols and positions
  center_index = large_index[nt]
  center_pos = [[size // 2, size // 2]]

  have_small = np.random.randint(3)
  ns = np.random.randint(7, 20)
  peri_index = np.random.permutation(len(small))[:ns]

  t = np.linspace(0, 2 * np.pi, ns, endpoint=False)
  peri_pos = 0.8 * die_radius * np.array([np.cos(t), np.sin(t)]) - size // 2
  peri_pos = list(peri_pos.T.astype(int))

  for nd in range(n_dies_per_type):
    # randomize symbols
    S_small = randomize_symbols(small[peri_index], 5, 12)
    S_large = randomize_symbols([large[center_index]], 5, 10)

    # create die and carve
    die = Die(size, radius=die_radius, thickness=500)

    if have_small:
      die.carve(S_small,
                peri_pos,
                strength=0.5,
                rotation=0,
                scaling=0.4)
      die.carve(S_large,
                center_pos,
                strength=1,
                rotation=0,
                scaling=1.2)
    else:
      die.carve(S_large,
                center_pos,
                strength=1,
                rotation=0,
                scaling=1.5)

    # strike coins
    for nc in range(n_coins_per_die):
      coin = Coin(size, radius=coin_radius, thickness=500)
      coin.generate_blank(smooth=5)
      tex_ind = np.random.randint(len(textures))
      tex = textures[tex_ind]
      coin.load_texture(tex)
      force = np.random.rand() * 0.5 + 0.3
      coin.strike(die, a=force, tilt=50, wearing_out=2)

      # erode
      if np.random.randint(3):
        n_iter = np.random.randint(3, 8)
        strength = np.random.randint(5, 12)
        coin.wear(strength=strength, part=True, radius=400, n_iter=n_iter)
      coin.wear(part=False, texture_strength=tex_coeffs[tex_ind])
      coin_database[nt][nd][nc] = coin
      update_progress((nt * n_dies_per_type * n_coins_per_die + nd * n_coins_per_die + nc) / total_coins, message="")

update_progress(1, message="")

# save
# dump(coin_database, datapath + '/batches/' + name + ".jbl")

###################################################
# render images
# color = (150 / 255, 150 / 255, 170 / 255)
# azimuth, slant = 0, np.pi / 4

# for nt in range(n_types):
#   fig, ax = plt.subplots(n_dies_per_type, n_coins_per_die)
#   for nd in range(n_dies_per_type):
#     for nc in range(n_coins_per_die):
#       coin = coin_database[nt][nd][nc]
#       I = coin.take_a_picture(azimuth, slant, shot_noise=0.1, color=color)
#       I_save = I * np.array([coin.mask, coin.mask, coin.mask]).T
#       ax[nd, nc].imshow(I)
#       ax[nd, nc].axis('off')

# plt.show()
