import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import os
from pathlib import Path
import io
from skimage.io import imread, imsave
import glob
from factory import *
from skimage.filters import gaussian
from skimage.transform import rescale, rotate

from skimage.transform import PiecewiseAffineTransform, warp


def random_nonrigid_transform(img, knots, intensity):
    rows, cols = img.shape[0], img.shape[1]

    src_cols = np.linspace(0, cols, knots, endpoint=True)
    src_rows = np.linspace(0, rows, knots, endpoint=True)
    src_rows, src_cols = np.meshgrid(src_rows, src_cols)
    src = np.dstack([src_cols.flat, src_rows.flat])[0]

    # add random perturbation
    dst_rows = src[:, 1] + (np.random.randn(src.shape[0]) - 0.5) * intensity
    dst_cols = src[:, 0] + (np.random.randn(src.shape[0]) - 0.5) * intensity
    dst = np.vstack([dst_cols, dst_rows]).T

    # calculate average displacement
    d = np.mean(np.sqrt((dst - src)**2))

    tform = PiecewiseAffineTransform()
    tform.estimate(dst, src)

    out = warp(img, tform, output_shape=img.shape)

    plt.figure()
    plt.imshow(img.max() - img, 'gray')
    plt.scatter(src[:, 0], src[:, 1], c='crimson', s=30)
    plt.xlim(-50, 300)
    plt.ylim(300, -50)
    plt.savefig("C:/Users/flarde01/Documents/L3i/Thèse Git/Manuscript/images/Chapter 1/PAT1.png", dpi=200)

    plt.figure()
    plt.imshow(-out, 'gray')
    plt.scatter(dst_cols, dst_rows, c='crimson', s=30, zorder=2)
    plt.scatter(src[:, 0], src[:, 1], facecolors='none', edgecolors='crimson', s=30, zorder=2, alpha=0.7)
    plt.xlim(-50, 300)
    plt.ylim(300, -50)

    for i in range(len(src)):
        U = dst_cols[i] - src[i, 0]
        V = dst_rows[i] - src[i, 1]
        plt.arrow(src[i, 0], src[i, 1], 0.7 * U, 0.7 * V, color='g', head_width=5, head_length=5, zorder=5)
    # plt.savefig("C:/Users/flarde01/Documents/L3i/Thèse Git/Manuscript/images/Chapter 1/PAT2.png", dpi=200)
    plt.show()
    return out, dst, d


a = np.random.randn(100)
print(np.std(a))

path = Path.cwd()
datapath = str(path) + '/'

# coin parameters
size = 1024
coin_radius = size // 2
die_radius = size // 2 - 100

# load symbols
small = []
for symbol in glob.glob(datapath + 'symbols/small/*.png')[9:]:
    S = imread(symbol, 1)
    S, _, _ = random_nonrigid_transform(S, 5, 10)
    small.append(S)
small = np.array(small)
center = []
for symbol in glob.glob(datapath + 'symbols/large/*.png'):
    S = imread(symbol, 1)
    S, _, _ = random_nonrigid_transform(S, 5, 5)
    center.append(S)

# generate random texture
texture = imread(datapath + 'textures/' + 't4.png', 1)
texture = normalize(texture.astype(float))

# initialize
die = Die(size, radius=die_radius, thickness=500)
coin = Coin(size, radius=coin_radius, thickness=500)
coin.load_texture(texture)

# choose symbols and positions
ns = 20
peri_index = np.random.permutation(len(small))[:ns]
center_index = 15

t = np.linspace(0, 2 * np.pi, ns, endpoint=False)
peri_pos = 0.8 * die_radius * np.array([np.cos(t), np.sin(t)]) - size // 2
peri_pos = list(peri_pos.T.astype(int))
center_pos = [[size // 2, size // 2]]

# carve
die.carve([center[center_index]],
          center_pos,
          strength=1,
          rotation=0,
          scaling=1.1)

# die.carve(small[peri_index],
#           peri_pos,
#           strength=0.5,
#           rotation=0,
#           scaling=0.3)

n = 1
for i in range(n):
    # strike the blank
    coin.generate_blank(smooth=5)
    coin.strike(die, a=0.5, wearing_out=2)

    # erode
    coin.wear(strength=20, part=True, radius=400, n_iter=1)

    coin.wear(part=False, texture_strength=0.05)

    # render image
    color = (150 / 255, 150 / 255, 170 / 255)
    azimuth, slant = 0, np.pi / 4

    I = coin.take_a_picture(azimuth, slant, shot_noise=0.1, color=color)
    I_save = I * np.array([coin.mask, coin.mask, coin.mask]).T
    # imsave(datapath + '/man_profile.png', (255 * I_save).astype(np.uint8))

    plt.figure(figsize=(10, 5))
    plt.subplot('121')
    plt.imshow(coin.piece * coin.mask, 'gray')
    plt.subplot('122')
    if color is not None:
        plt.imshow(I)
    else:
        plt.imshow(I * coin.mask, 'gray')
plt.show()
