# Libraries
from IPython.display import clear_output
import numpy as np
import cv2
from skimage.morphology import skeletonize
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.patches import Circle
from matplotlib.colors import hsv_to_rgb
import os
import sys
import glob
from fastaniso import *
from skimage import exposure
from skimage.filters import threshold_local
from skimage.morphology import closing, disk, erosion, label, dilation
from skimage.transform import resize, rotate, rescale
from skimage.transform import PiecewiseAffineTransform, warp
from skimage.transform import pyramid_gaussian
from skimage.util.shape import view_as_windows
from skimage import data
from skimage.feature import register_translation
from skimage.util import random_noise
from skimage.measure import regionprops
import scipy as sp
from scipy.ndimage.morphology import binary_fill_holes
from scipy import ndimage

from scipy.signal import convolve2d
from skimage.filters import gaussian
from skimage.filters.rank import entropy
from skimage.morphology import disk

from skimage.feature import peak_local_max
from scipy.ndimage import maximum_filter

from skimage.measure import label, regionprops
from skimage.color import label2rgb

from skimage.feature import hog
from skimage import exposure

from scipy.stats import gaussian_kde

from utils import *

import imutils


def preprocess(img, s):

    img = cv2.resize(img, (s, s))
    img = exposure.rescale_intensity(img)
    _, mask = entropyseg(img)

    mask = np.array(255 * mask, dtype=np.uint8)
    _, contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    bbox = cv2.boundingRect(contours[0])
    top_left, bottom_right = (bbox[0], bbox[1]), (bbox[0] + bbox[2], bbox[1] + bbox[3])
    cropped = img[bbox[1]:bbox[1] + bbox[3], bbox[0]:bbox[0] + bbox[2]]
    cropped = cv2.resize(cropped, (s, s))  # sloppy, but whatever
    return cropped


def cutByBlock(img, blockSize=32, blockStride=16):
    # img must me squared, power of 2
    blocks = []
    n_blocks = int(img.shape[0] / blockStride)
    for i in range(n_blocks - 1):
        for j in range(n_blocks - 1):
            block = img[i * blockStride:i * blockStride + blockSize,
                        j * blockStride:j * blockStride + blockSize]
            blocks.append(block)

    return blocks


def MLEM(images, slant=np.pi / 4, withangles=False, angles=1, strength=20, verbose=0):
    """ Compute the 'angular gradient' = dI/dtheta from a set of images"""
    """ Here we compute everything in one function. No scalespaces involved"""
    """ Only tilts ! """
    h = images[0].shape[0]
    w = images[0].shape[1]

    N = len(images)
    dtheta = 2 * np.pi / N

    S = np.zeros((h, w))
    A = np.zeros((h, w))
    gmax = np.zeros((h, w))

    for i in range(N):
        if verbose > 0:
            print("Computing MLEM {}/{}".format(i + 1, N))
        if i == N - 1:
            I1 = anisodiff(images[i], strength)
            I2 = anisodiff(images[0], strength)
        else:
            I1 = anisodiff(images[i], strength)
            I2 = anisodiff(images[i + 1], strength)

        # magnitude
        Si = ((I2 - I1) / dtheta)**2
        S += Si

        # angles
        if withangles:
            for x in range(h):
                for y in range(w):
                    g = I1[x, y]
                    if g > gmax[x, y]:
                        A[x, y] = angles[i]
                        gmax[x, y] = I1[x, y]

    g = S

    E_mag = np.sqrt(g)
    E_mag = normalize(E_mag, uint8=True)

    return E_mag, A

def MLEM_for_HOME(images, strength=5, verbose=0):

    h = images[0].shape[0]
    w = images[0].shape[1]

    N = len(images)
    dtheta = 2 * np.pi / N

    S = np.zeros((h, w))
    gmax = np.zeros((h, w))

    for i in range(N):
        if verbose > 0:
            print("Computing MLEM {}/{}".format(i + 1, N))
        if i == N - 1:
            I1 = anisodiff(images[i], strength)
            I2 = anisodiff(images[0], strength)
        else:
            I1 = anisodiff(images[i], strength)
            I2 = anisodiff(images[i + 1], strength)

        # magnitude
        Si = ((I2 - I1) / dtheta)**2
        S += Si

    # angles
    A = [gaussian(images[i], 3) for i in range(len(images))]

    g = S

    E_mag = np.sqrt(g)
    E_mag = normalize(E_mag, uint8=True)

    return E_mag, np.array(A).T

def preprocess(img, s, crop=True):
    img = cv2.resize(img, (s, s))
    img = exposure.rescale_intensity(img)

    if crop:
        _, mask = entropyseg(img)
        mask = np.array(255 * mask, dtype=np.uint8)
        _, contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        bbox = cv2.boundingRect(contours[0])
        top_left, bottom_right = (bbox[0], bbox[1]), (bbox[0] + bbox[2], bbox[1] + bbox[3])
        cropped = img[bbox[1]:bbox[1] + bbox[3], bbox[0]:bbox[0] + bbox[2]]
        cropped = cv2.resize(cropped, (s, s))  # sloppy, but whatever
        return cropped

    else:
        return img


def adaptive_threshold(image, block_size, offset=10):
    binary_adaptive = threshold_local(image, block_size, offset=10) < image
    binary_adaptive = binary_adaptive / 255
    edgeblobs = 1 - (binary_adaptive > 0)

    # contours of blobs
    contours = sp.ndimage.distance_transform_edt(edgeblobs)
    contours[contours != 1] = 0

    # Small islands elimination (i.e. holes) - using a hole filling function leads to non relevant results (ex : 'D' or 'O' shapes)
    cont_l = label(1 - contours, 4)  # labelled regions without contours : let appear small island (conn 4)
    regions = regionprops(cont_l)
    small_regions = np.zeros(edgeblobs.shape)
    for region in regions:
        if region.area < 70 / 2:  # what threshold ?
            small_regions[region.coords[:, 0], region.coords[:, 1]] = 1  # keep small regions

    edgeblobs = edgeblobs = (small_regions + edgeblobs) > 0

    cont_l = label(edgeblobs, 4)  # labelled regions without contours : let appear small island (conn 4)
    regions = regionprops(cont_l)
    big_regions = np.zeros(edgeblobs.shape)
    for region in regions:
        if region.area > 50:  # what threshold ?
            big_regions[region.coords[:, 0], region.coords[:, 1]] = 1  # keep small regions

    return edgeblobs, big_regions


def normalize(E, uint8=False):
    E_min = E.min()
    E_max = E.max()
    if not uint8:
        return np.array((E - E_min) / (E_max - E_min), dtype=float)
    else:
        return np.array(255 * (E - E_min) / (E_max - E_min), dtype=np.uint8)

def preprocess_mlem(mlem, mask, shape=(1024,1024)):
    L = label(mask>0)
    regions = regionprops(L)
    ind_max = np.argsort([r.area for r in regions])[-1]
    x, y = regions[ind_max].coords.T

    cropped = mlem[min(x):max(x), min(y):max(y)]
    mask = mask[min(x):max(x), min(y):max(y)]
    if shape is not None:
        res = resize(cropped*mask, shape)
    else:
        res = cropped*mask
    return res

def probabilityMap(g, kernel_size=15):
    g = np.array(g, dtype=np.float)
    kde = gaussian(g, kernel_size)
    kde = normalize(kde, uint8=True)
    return kde

def feature_points(mlem, mask, probability_map, min_distance, threshold_min):
    # extract feature points
    coordinates = extract_coordinates(probability_map, mask, min_distance=min_distance, threshold_min=threshold_min)
    return coordinates

def extract_coordinates(kde_map, mask, min_distance=10, threshold_min=50):
    mask = np.array(mask == 0, dtype=bool)
    coords_img = peak_local_max(kde_map,
                                min_distance=min_distance,
                                threshold_abs=threshold_min,
                                indices=False,
                                exclude_border=True)

    label_image = label(coords_img)

    centers = []
    for region in regionprops(label_image):
        coords = region.coords
        center = np.array([int(np.mean(coords[:, 0])), int(np.mean(coords[:, 1]))])

        if mask[np.mod(int(center[0]), mask.shape[0]-1), np.mod(int(center[1]), mask.shape[1]-1)]:
            continue
        
        centers.append(center)
    coordinates = np.array(centers)

    return coordinates

def geometric_coherence(E1, E2, pts1, pts2, tol=0.01, register=False):
    if register:
        E2 = resize(E2, E1.shape)
        t, _, _ = register_translation(E2,E1)
        pts2 = pts2 + t
    else:
        t = -1
    
    r = tol * E1.shape[0]
    
    matches = []
    
    for p1 in pts1:
        for p2 in pts2:
            if np.linalg.norm(p1-p2) <= r:
                matches.append([p1, p2])
    
    return np.array(matches), t

def gaussian_window_mask(window, sigma):
    x = np.arange(window.shape[0])
    y = np.arange(window.shape[1])
    p = np.array([[xx,yy] for xx in x for yy in y])
    c = np.array([len(x)/2, len(y)/2])

    g = np.exp(-(p[:,0]-c[0])**2/(2*sigma**2))*np.exp(-(p[:,1]-c[1])**2/(2*sigma**2))
    g = g.reshape(window.shape)

    return g

def FP_maxima_gaussian(E, A, kernel_size=8, distance_min=10, threshold_min=0.2):
    P = probabilityMap(E, kernel_size=kernel_size)
    P = normalize(P)
    fp = feature_points(E, E>0, P, min_distance=distance_min, threshold_min=threshold_min)
    return fp

def FP_maxima_entropy(E, A, kernel_size=8, distance_min=16, threshold_min=0.8):
    H = entropy(normalize(E), disk(kernel_size))
    H = normalize(H)
    fp = feature_points(E, E>0, H, min_distance=distance_min, threshold_min=threshold_min)
    return fp

def FP_maxima_gaussian_scales(E, A, kernel_size=16, distance_min=16, threshold_min=0.2, max_layer=4, tol=32, display=False):
    pyramid = tuple(pyramid_gaussian(E, max_layer=max_layer, downscale=2))
    
    fps = []
    if display:
        fig, ax = plt.subplots(1, len(pyramid), figsize=(20,10))
    for i in range(len(pyramid)):
        fp = FP_maxima_gaussian(pyramid[i], -1, int(kernel_size/2**i), int(distance_min/2**i), threshold_min)
        fps.append(fp * 2**i) # rescale the points
        if display:
            ax[i].imshow(pyramid[i], 'gray')
            ax[i].scatter(fp[:,1], fp[:,0], s=50, c='r')
            ax[i].axis('off')
    fps = np.array(fps)
    
    if display:
        plt.figure(figsize=(7,7))
        colors = hsv_to_rgb([[np.random.rand(), 1, 1] for i in range(len(fps))])
        for i in range(len(fps)):
            plt.scatter(fps[i][:,0], fps[i][:,1], s=300, marker='x', c=colors[i])
            plt.gca().invert_yaxis()
            plt.axis('off')

    # count the number of scales that include each order 0 maxima within a circle of radius 'tol'
    count = {j:0 for j in range(len(fps[0]))} # counting per point of scale 0
    for j in range(len(fps[0])):
        pt = fps[0][j]
        if display:
            C = Circle(pt, tol, fill=False, edgecolor=colors[0], lw=1)
            plt.gca().add_patch(C)
        for scale in range(1, len(fps)):
            for k in range(len(fps[scale])):
                pt_s_k = fps[scale][k]
                if np.linalg.norm(pt-pt_s_k) <= tol:
                    count[j] += 1
                    break # no need to check the rest
    if display:
        plt.show()
    # keep only if the point is retrieved at all subscales
    n_subscales = len(fps) - 1
    fp = fps[0][np.argwhere(np.array(list(count.values())) == n_subscales).T[0]]
    return fp

def FP_maxima_entropy_scales(E, A, kernel_size=32, distance_min=32, threshold_min=0.8, max_layer=4, tol=32, display=False):
    pyramid = tuple(pyramid_gaussian(E, max_layer=max_layer, downscale=2))

    fps = []
    if display:
        fig, ax = plt.subplots(1, len(pyramid), figsize=(20,10))
    for i in range(len(pyramid)):
        fp = FP_maxima_entropy(pyramid[i], -1, int(kernel_size/2**i), int(distance_min/2**i), threshold_min)
        fps.append(fp * 2**i) # rescale the points
        if display:
            ax[i].imshow(pyramid[i], 'gray')
            ax[i].scatter(fp[:,1], fp[:,0], s=50, c='r')
            ax[i].axis('off')
    fps = np.array(fps)
    
    if display:
        plt.figure(figsize=(7,7))
        colors = hsv_to_rgb([[np.random.rand(), 1, 1] for i in range(len(fps))])
        for i in range(len(fps)):
            plt.scatter(fps[i][:,0], fps[i][:,1], s=300, marker='x', c=colors[i])
            plt.gca().invert_yaxis()
            plt.axis('off')

    # count the number of scales that include each order 0 maxima within a circle of radius 'tol'
    count = {j:0 for j in range(len(fps[0]))} # counting per point of scale 0
    for j in range(len(fps[0])):
        pt = fps[0][j]
        if display:
            C = Circle(pt, tol, fill=False, edgecolor=colors[0], lw=1)
            plt.gca().add_patch(C)
        for scale in range(1, len(fps)):
            for k in range(len(fps[scale])):
                pt_s_k = fps[scale][k]
                if np.linalg.norm(pt-pt_s_k) <= tol:
                    count[j] += 1
                    break # no need to check the rest
    if display:
        plt.show()
        
    # keep only if the point is retrieved at all subscales
    n_subscales = len(fps) - 1
    fp = fps[0][np.argwhere(np.array(list(count.values())) == n_subscales).T[0]]
    return fp

def FP_centroids_blobs(E, A, threshold_type='fixed', block_size=101, offset=-30):
    if threshold_type in ['absolute', 'fixed']:
        intensity = E[np.where(E>0)]
        mu = np.mean(intensity)
        sigma = np.std(intensity)
        tau = mu + sigma

        blobmap = E > tau

    elif threshold_type in ['adaptive', 'local']:
        blobmap = cv2.adaptiveThreshold((E*255).astype(np.uint8), 
                                        255, 
                                        cv2.ADAPTIVE_THRESH_GAUSSIAN_C, 
                                        cv2.THRESH_BINARY, 101, -30)
    blobmap = clean_blobs(blobmap)
    L = label(blobmap)

    fp = []
    regions = regionprops(L)
    for reg in regions:
        fp.append(reg.centroid)
    fp = np.array(fp)
    return fp

def FP_skeleton(E, A, threshold_type='fixed', block_size=101, offset=-30):
    if threshold_type in ['absolute', 'fixed']:
        intensity = E[np.where(E>0)]
        mu = np.mean(intensity)
        sigma = np.std(intensity)
        tau = mu + sigma
        blobmap = E > tau

    elif threshold_type in ['adaptive', 'local']:
        blobmap = cv2.adaptiveThreshold((E*255).astype(np.uint8), 
                                        255, 
                                        cv2.ADAPTIVE_THRESH_GAUSSIAN_C, 
                                        cv2.THRESH_BINARY, 101, -30)
    blobmap = clean_blobs(blobmap)
    skel = skeletonize(blobmap>0)

    fp = np.argwhere(skel)
    return fp

def FP_Harris(E, A, mode='union',
              angle_type='mean', 
              max_layer=2, 
              gauss_kernel_size=16, distance_min=16, threshold_min=0.2, 
              w_size_oct=3,
              response_type=0, k=0.05,
              cross_scale_radius=8, n_subscales=2,
             display=False):
    
    E = normalize(E)
    # calculate angles
    if angle_type in ['max']:
        loc = np.argmax(A, axis=2).T
        angles = 2 * np.pi * loc / loc.max()
        
    elif angle_type in ['average', 'mean']:
        theta = np.linspace(0, 2*np.pi, A.shape[2])
        vect = []
        for i in range(len(theta)):
            t = theta[i]
            a = A[:,:,i]
            v = np.array([a*np.cos(t), a*np.sin(t)])
            vect.append(v)
        vect = np.array(vect)
        mean_vect = np.mean(vect, axis=0)
        angles = np.arctan2(mean_vect[1],mean_vect[0]).T
        angles = np.mod(angles, 2*np.pi)

    # back to cartesian
    Ex = E * np.cos(angles)
    Ey = E * np.sin(angles)
    
    pyramid_x = tuple(pyramid_gaussian(Ex, max_layer=max_layer, downscale=2))
    pyramid_y = tuple(pyramid_gaussian(Ey, max_layer=max_layer, downscale=2))

    w = np.array(2*[2**w_size_oct]) # w = 8 corresponds interestingly to "lambda", the average ridge thickness.
    fps = []
    if display:
        fig, ax = plt.subplots(1, max_layer+1, figsize=(5*(max_layer+1), 5))
    for octave in range(len(pyramid_x)):
        print('Octave {}'.format(octave))
        Ex_oct = pyramid_x[octave]
        Ey_oct = pyramid_y[octave]

        # windows
        print('Computing tensor...')
        Wx = view_as_windows(Ex_oct, w)
        Wy = view_as_windows(Ey_oct, w)

        # gaussian window
        Wx = Wx * gaussian_window_mask(Wx[0,0], w[0]/2+1)
        Wy = Wy * gaussian_window_mask(Wy[0,0], w[0]/2+1)

        # tensor components
        Ix2 = np.sum(Wx**2, axis=(2,3))
        Iy2 = np.sum(Wy**2, axis=(2,3))
        Ixy = np.sum(Wx*Wy, axis=(2,3))

        M = np.array([[Ix2, Ixy], [Ixy, Iy2]])
        M = np.swapaxes(M, 1, 3)
        M = np.swapaxes(M, 0, 2)

        # Harris response
        print('Computing eigenvalues...')
        eigs = np.zeros(M.shape[:2])
        for i in range(M.shape[0]):
            for j in range(M.shape[1]):
                if response_type == 0:
                    eig_value = np.linalg.det(M[i,j]) / np.trace(M[i,j]) # stable
                else:
                    eig_value = np.linalg.det(M[i,j]) - k*np.trace(M[i,j])**2 # requires painful tuning
                eigs[i,j] = eig_value
        eigs[np.isinf(eigs)] = 0
        eigs[np.isnan(eigs)] = 0
        eigs = normalize(eigs)

        if display:
            ax[octave].imshow(eigs, 'pink')

        fp = FP_maxima_gaussian(eigs, -1, int(gauss_kernel_size/2**octave), int(distance_min/2**octave), threshold_min)
        fps.append(2**octave*fp) # rescale

        if display:
            ax[octave].imshow(eigs, 'pink')
            try:
                ax[octave].scatter(fp[:,1], fp[:,0], c='r', s=50)
            except:
                print('No fp , octave:', octave)
        clear_output(wait=True)
        
    if display:
        fig, ax = plt.subplots(1,3,figsize=(15,5))
        ax[0].imshow(angles, 'hsv')
        ax[0].set_title('Angles', color='w', fontsize=18)
        ax[0].axis('off')
        ax[1].imshow(Ex, 'bone')
        ax[1].axis('off')
        ax[1].set_title('Gradient intensity x', color='w', fontsize=18)
        ax[2].imshow(Ey, 'bone')
        ax[2].axis('off')
        ax[2].set_title('Gradient intensity y', color='w', fontsize=18)

    fps = np.array(fps)
    
    if display:
        plt.figure(figsize=(5,5))
        colors = hsv_to_rgb([[np.random.rand(), 1, 1] for i in range(len(fps))])
        for i in range(len(fps)):
            try:
                plt.scatter(fps[i][:,0], fps[i][:,1], s=200, marker='x', c=colors[i])
                plt.gca().invert_yaxis()
                plt.axis('off')
            except:
                print('No fp :', i)
    
    print('Non-maxima suppression...')
    if mode in ['inter']:
        # Intersection - only keep points that are present within a certain amount of subscales (octaves).
        # The higher the amount of subscales, the more robust the feature points are, in theory.
        # count the number of scales that include each order 0 maxima within a circle of radius 'tol'
        count = {j:0 for j in range(len(fps[0]))} # counting per point of scale 0
        for j in range(len(fps[0])):
            pt = fps[0][j]

            for scale in range(1, len(fps)):
                for k in range(len(fps[scale])):
                    pt_s_k = fps[scale][k]
                    if np.linalg.norm(pt-pt_s_k) <= cross_scale_radius:
                        count[j] += 1
                        break # no need to check the rest

        fp = fps[0][np.argwhere(np.array(list(count.values())) >= n_subscales).T[0]]
        
    elif mode in ['union']:
        # Union - keep all the points across the octaves.
        # The points are however gathered if too close across octaves.
        mask = np.zeros(E.shape, dtype=bool)
        e = disk(cross_scale_radius)
        masks = []

        # create a mask for the tolerance - dilated sum of all octave point masks
        for fp in fps:
            mask = np.zeros(E.shape)
            try:
                mask[fp[:,0], fp[:,1]] = 1
                masks.append(mask)
            except:
                pass
        if len(masks) > 0:
            total = np.sum(masks, axis=0) > 0
            dilated = dilation(total, e)
            L = label(dilated)

            # for all feature points detected, keep the one corresponding to the highest octave
            # lower octave lose precision (locality).
            # The tolerance is defined as the presence of several points within the tolerance mask
            fp = []
            for l in range(1, L.max()):
                b = True
                i = 0
                while b:
                    I = (L == l) * masks[i]
                    loc = np.array(np.where(I)).T
                    b = len(loc) == 0
                    i += 1
                fp.append(loc[0])
        else:
            fp = [[-1,-1]]
            
    if display:
        plt.show()
    fp = np.array(fp)
    return fp

def FP_random(E, A, n):
    h, w = E.shape
    x = np.random.randint(0, h, (n,1)).T[0]
    y = np.random.randint(0, w, (n,1)).T[0]
    fp = np.array(list(zip(x,y)))
    fp = np.array([[x,y] for (x,y) in fp if E[x,y]>0])
    return fp

def show_fp(img, fp, figsize=(10,10), fontcolor='w'):
    fig, ax = plt.subplots(1,1,figsize=figsize)
    ax.imshow(img, 'gray')
    ax.scatter(fp[:,1], fp[:,0], c='r', s=50)
    ax.axis('off')
    ax.set_title('{} points'.format(len(fp)), color=fontcolor)
    return ax

def rotate_points(points, center, angle):
    ox, oy = center
    rpoints = []
    for pt in points:
        px, py = pt
        qx = ox + np.cos(angle) * (px - ox) - np.sin(angle) * (py - oy)
        qy = oy + np.sin(angle) * (px - ox) + np.cos(angle) * (py - oy)
        rpoints.append([qx, qy])
    return np.array(rpoints)

def random_nonrigid_transform(img, density, intensity):
    rows, cols = img.shape[0], img.shape[1]

    src_cols = np.linspace(0, cols, int(cols*density))
    src_rows = np.linspace(0, rows, int(rows*density))
    src_rows, src_cols = np.meshgrid(src_rows, src_cols)
    src = np.dstack([src_cols.flat, src_rows.flat])[0]

    # add random perturbation
    dst_rows = src[:, 1] + (np.random.randn(src.shape[0])-0.5)*intensity
    dst_cols = src[:, 0] + (np.random.randn(src.shape[0])-0.5)*intensity
    dst = np.vstack([dst_cols, dst_rows]).T
    
    # calculate average displacement
    d = np.mean(np.sqrt((dst-src)**2))

    tform = PiecewiseAffineTransform()
    tform.estimate(src, dst)

    out = warp(img, tform, output_shape=img.shape)
    return out, dst, d

def erode(I, patch_size, x, y):
    slice_ = np.s_[x-patch_size:x+patch_size,y-patch_size:y+patch_size]
    window = np.zeros((2*patch_size, 2*patch_size))
    gmask = gaussian_window_mask(window, patch_size/2)
    I[slice_] = gaussian(I[slice_], patch_size//4) * gmask + I[slice_] * (1-gmask)
    return normalize(I)

def stability(matches, pts1):
    if len(matches) == 0 or len(pts1) == 0:
        return 0
    n = len(pts1)
    m = len(unique2D(matches[:,0]))
    return m/n

def precision(matches, pts1):
    if len(matches) == 0 or len(pts1) == 0:
        return 1
        
    n = len(pts1)

    points = matches[:,0]
    points_ = unique2D(points)
    
    count = np.zeros(len(points_))
    for p in points:
        b, inds = belong_to_list(p, points_)
        if b:
            for i in inds:
                count[i] += 1
    count = count > 1
            
    prec_score = 1 - count.sum() / n
    return prec_score
    
def repeatability(imgs1, imgs2, M1, M2, method, perturbation, tol, params, *args):
    stab, prec = [], []
    n = len(params)
    
    if perturbation not in ['random']:
        E1, A1 = MLEM_for_HOME(imgs1, strength=1, verbose=0)
        A1 = resize(A1, (1024, 1024, A1.shape[2]))
        M1 = resize(M1, (1024, 1024))
        E1 = resize(E1, (1024, 1024))
        E1 = preprocess_mlem(E1, M1, (1024,1024))
    
        pts1 = method(E1, A1, *args[0])
    
#     plt.figure(figsize=(7,7))
#     plt.imshow(E1, 'gray')
#     plt.scatter(pts1[:,1], pts1[:,0], s=50, c='r')
#     plt.title('E1')
#     plt.show()
    
    if perturbation in ['rotate', 'rotation']:
        for i, theta in enumerate(params):
            imgs2_r = []
            for I in imgs2:
                I = rotate(I, theta*180/np.pi)
                imgs2_r.append(I)
            
            E2, A2 = MLEM_for_HOME(imgs2_r, strength=1, verbose=0)
            M2 = resize(M2, (1024, 1024))
            M2 = rotate(M2, theta*180/np.pi)
            E2 = preprocess_mlem(E2, M2)
            
            pts2 = method(E2, A2, *args[0])
            pts2 = rotate_points(pts2, (E2.shape[0] / 2 - 0.5, E2.shape[1] / 2 - 0.5), -theta)
            
            try:
                m, _ = geometric_coherence(E1, E2, pts1, pts2, tol=tol)
            except:
                m = []

            st = stability(m, pts1)
            c = precision(m, pts1)
            stab.append(st)
            prec.append(c)
            
            update_progress(i / n, 
                             "Computing repeatability | rotation | theta = {}° | tol = {} px".format(np.round(theta*180/np.pi,2), np.round(tol*1024, 2)))
        update_progress(1, "Computing repeatability | rotation | finished")
    
    elif perturbation in ['scale', 'scaling']:
        for i, s in enumerate(params):
            imgs2_s = []
            for I in imgs2:
                I = resize(I, (1024,1024))
                I = rescale(I, s)
                imgs2_s.append(I)
                
            E2, A2 = MLEM_for_HOME(imgs2_s, strength=1, verbose=0)
            M2 = resize(M2, E2.shape)
            E2 = preprocess_mlem(E2, M2, shape=None)
            A2 = resize(A2, (E2.shape[1], E2.shape[0], A2.shape[2]))
            
            pts2 = method(E2, A2, *args[0])
            pts2 = pts2 / s

            try:
                m, _ = geometric_coherence(E1, E2, pts1, pts2, tol=tol, register=False)
            except:
                m = []

            st = stability(m, pts1)
            c = precision(m, pts1)
            stab.append(st)
            prec.append(c)
            
            update_progress(i / n, 
                             "Computing repeatability | scaling | factor = {} | tol = {} px".format(np.round(s,4), np.round(tol*1024,2)))
        update_progress(1, "Computing repeatability | scaling | finished")
        
    elif perturbation in ['noise']:
        for i, u in enumerate(params):
            imgs2_u = []
            for I in imgs2:
                I = random_noise(I, mode='gaussian', var=u)
                imgs2_u.append(I)
                
            E2, A2 = MLEM_for_HOME(imgs2_u, strength=1, verbose=0)
            M2 = resize(M2, (1024, 1024))
            E2 = preprocess_mlem(E2, M2)
            
            pts2 = method(E2, A2, *args[0])
            
#             plt.figure(figsize=(7,7))
#             plt.imshow(E2, 'gray')
#             plt.scatter(pts2[:,1], pts2[:,0], s=50, c='r')
#             plt.title('E2')
#             plt.show()

            try:
                m, _ = geometric_coherence(E1, E2, pts1, pts2, tol=tol)
            except:
                m = []

            st = stability(m, pts1)
            c = precision(m, pts1)
            stab.append(st)
            prec.append(c)
            
            update_progress(i / n, 
                             "Computing repeatability | noise | sigma = {} | | tol = {} px".format(np.round(u,4), np.round(tol*1024, 2)))
        update_progress(1, "Computing repeatability | noise | finished")
    
    elif perturbation in ['random']:
        imgs2_u = []
        for I in imgs2:
            imgs2_u.append(I)

        E1, A1 = MLEM_for_HOME(imgs1, strength=1, verbose=0)
        M1 = resize(M1, (1024, 1024))
        E1 = resize(E1, (1024, 1024))
        E1 = preprocess_mlem(E1, M1, (1024,1024))

        E2, A2 = MLEM_for_HOME(imgs2_u, strength=1, verbose=0)
        M2 = resize(M2, (1024, 1024))
        E2 = preprocess_mlem(E2, M2)
        
        for i, u in enumerate(params):
            st, c = 0, 0
            pts1 = FP_random(E1, A1, u)
            for rep in range(100): # ensure convergence
                pts2 = FP_random(E2, A2, u)
                try:
                    m, _ = geometric_coherence(E1, E2, pts1, pts2, tol=tol)
                except:
                    m = []
                st += stability(m, pts1)/100
                c += precision(m, pts1)/100
                
            stab.append(st)
            prec.append(c)
            
            update_progress(i / n, 
                             "Computing repeatability | random sampling | n points = {} | | tol = {} px".format(np.round(u,4), np.round(tol*1024, 2)))
        update_progress(1, "Computing repeatability | noise | finished")
    
    elif perturbation in ['erode', 'erosion']:
        h, w = imgs2[0].shape
        for i, size in enumerate(params):
            st = 0
            c = 0
            for k in range(10):
                x = np.random.randint(size, h-size)
                y = np.random.randint(size, w-size)
                
                imgs2_s = []
                for I in imgs2:
                    Ir = I.copy()
                    imgs2_s.append(erode(Ir, size, x, y))
                    
                E2, A2 = MLEM_for_HOME(imgs2_s, strength=1, verbose=0)
                M2 = resize(M2, (1024, 1024))
                E2 = preprocess_mlem(E2, M2)
                
                pts2 = method(E2, A2, *args[0])

                try:
                    m, _ = geometric_coherence(E1, E2, pts1, pts2, tol=tol)
                except:
                    m = []

                st += stability(m, pts1)
                c += precision(m, pts1)
            stab.append(st/10)
            prec.append(c/10)
            
            update_progress(i / n, 
                             "Computing repeatability | erosion | patch size = {}x{} px | tol = {}".format(np.round(2*size,4), np.round(2*size,4), np.round(tol*1024, 2)))
            
    elif perturbation in ['random_piecewise_affine']:
        for i, u in enumerate(params):
            density = 0.02
            imgs2_u = []
            for I in imgs2:
                I, _, _ = random_nonrigid_transform(I, density, u)
                imgs2_u.append(I)
                
            E2, A2 = MLEM_for_HOME(imgs2_u, strength=1, verbose=0)
            M2 = resize(M2, (1024, 1024))
            E2 = resize_mlem(E2, M2)
            
            pts2 = method(E2, A2, *args[0])
            
            try:
                m, t = geometric_coherence(E1, E2, pts1, pts2, tol=tol)
            except:
                m = []

            st = stability(m, pts1)
            c = precision(m, pts1)
            stab.append(st)
            prec.append(c)
            
            update_progress(i / n, 
                             "Computing repeatability | random piecewise affine | sigma = {} | tol = {}".format(np.round(u,4), np.round(tol*1024, 2)))
        update_progress(1, "Computing repeatability | random piecewise affine | finished")
        
        img = E1
        
    return np.array(stab), np.array(prec)


def getPatches(img, mlem, mask, coordinates, window_size, patch_size=64, nzeromin=0.5, verbose=0):
    patches = {c: [] for c in range(len(coordinates))}
    coords = {c: [] for c in range(len(coordinates))}
    size = int((window_size - 1) / 2)
    
    for c, (x, y) in enumerate(coordinates):
        if verbose > 0:
            if c % 10 == 0:
                print("Computing patches {}/{}".format(c + 1, len(coordinates)))
        smin_x = x - size
        smax_x = x + size + 1
        smin_y = y - size
        smax_y = y + size + 1
        
        if smin_x < 0 or smax_x >= img.shape[0] or smin_y < 0 or smax_y >= img.shape[1]:
            patches.pop(c, None)
            coords.pop(c, None)
            continue

        patch = img[smin_x:smax_x, smin_y:smax_y]
        inMask = np.nonzero(mask[smin_x:smax_x, smin_y:smax_y])[0]
        if len(inMask) <= nzeromin * patch.shape[0]*patch.shape[1]:
            patches.pop(c, None)
            coords.pop(c, None)
            continue
        patches[c] = cv2.resize(patch, (patch_size, patch_size))
        coords[c] = (x,y)

    return np.array(list(patches.values())), np.array(list(coords.values()))

def getPatches_angles(A, mask, coordinates, window_size, path_size=64, verbose=0):
    patches = {c: [] for c in range(len(coordinates))}
    coords = {c: [] for c in range(len(coordinates))}
    size = int((window_size - 1) / 2)
    
    for c, (x, y) in enumerate(coordinates):
        if verbose > 0:
            if c % 10 == 0:
                print("Computing patches {}/{}".format(c + 1, len(coordinates)))
        smin_x = x - size
        smax_x = x + size + 1
        smin_y = y - size
        smax_y = y + size + 1
        
        if smin_x < 0 or smax_x >= A.shape[0] or smin_y < 0 or smax_y >= A.shape[1]:
            patches.pop(c, None)
            coords.pop(c, None)
            continue

        A_patch = A[smin_x:smax_x, smin_y:smax_y]
        A_patch_resized = []
        for apatch in A_patch.T:
            A_patch_resized.append(cv2.resize(apatch, (64, 64)))
        patches[c] = np.array(A_patch_resized).T
        coords[c] = (x,y)

    return np.array(list(patches.values()))


def sum_of_hog(imgs, orientations=8, pixels_per_cell=(8, 8), cells_per_block=(4, 4), resize=0):
    if resize > 0:
        res = np.zeros(imutils.resize(imgs[0], width=resize).shape[:2])
    else:
        res = np.zeros(imgs[0].shape[:2])
    feature_vector_created = False
    for i in range(len(imgs)):

        image = imgs[i]
        if resize > 0:
            image = imutils.resize(image, width=resize)
        fd, hog_image = hog(image, orientations=orientations, pixels_per_cell=pixels_per_cell,
                            cells_per_block=cells_per_block, visualize=True, multichannel=False,
                            feature_vector=True)

        if not feature_vector_created:
            v = np.zeros(len(fd))

        hog_image_rescaled = normalize(hog_image, uint8=False)
        res += hog_image_rescaled / len(imgs)
        v += np.array(fd) / len(imgs)

    return np.array(255 * res, dtype=np.uint8), v

class HOME:
    def __init__(self, n_cells=(8,8), block_size=(2,2), overlap=0.5, block_normalization=True, truncate=0):
        self.n_cells = n_cells
        self.block_size = block_size
        self.overlap = overlap
        self.bn = block_normalization
        self.t = truncate
        
    def compute(self, A_patch):
        no = A_patch.shape[2]
        # cells and blocks
        pixels_per_cell = (A_patch.shape[0] // self.n_cells[0], A_patch.shape[1] // self.n_cells[1])
        cells = np.array([A_patch[k * pixels_per_cell[0]:(k + 1) * pixels_per_cell[0],
                                  l * pixels_per_cell[1]:(l + 1) * pixels_per_cell[1]]
                          for k in range(self.n_cells[0])
                          for l in range(self.n_cells[1])])

        # cells binning
        H = []
        vis = []
        for c in cells:
            hist = np.zeros(no)
            for k in range(c.shape[0]):
                for l in range(c.shape[1]):
                    for i, v in enumerate(c[k, l]):
                        hist[i] += v if v > self.t else 0

            hist = np.array(hist)
            H_n = hist / np.linalg.norm(hist)
            H.append(hist)
            vis.append(H_n)
        H = np.array(H)

        if self.bn:
            # block normalization
            step_x = int(self.block_size[0] * (1 - self.overlap))
            step_y = int(self.block_size[1] * (1 - self.overlap))
            H_ = np.reshape(H, (self.n_cells[0], self.n_cells[1], no))
            total_vec = []
            for k in range(0, self.n_cells[0] - 1, step_x):
                for l in range(0, self.n_cells[1] - 1, step_y):
                    vec = []
                    for cx in range(k, k + 2):
                        for cy in range(l, l + 2):
                            vec += list(H_[cx, cy])
                    total_vec = total_vec + list(np.array(vec) / np.linalg.norm(vec))
            H = np.array(total_vec)
        return H.flatten(), vis
    
    def display(self, patch, hists, r=0.7, with_magnitude=False):
        if with_magnitude:
            mag = normalize(gaussian(patch, 8))
        c = np.arange(self.n_cells[0] * self.n_cells[1])
        x, y = self.n_cells[0] * (c // self.n_cells[0]) + self.n_cells[0] // 2 - 0.5, self.n_cells[1] * (c % self.n_cells[1]) + self.n_cells[1] // 2 - 0.5
        x = x * patch.shape[0] / len(c)
        y = y * patch.shape[1] / len(c)

        r = patch.shape[0] // self.n_cells[0] * r

        for j, hist in enumerate(hists):
            no = len(hist)
            lines = np.array([[hist[i] * r * np.cos(2 * i * np.pi / no) + x[j], hist[i] * r * np.sin(2 * i * np.pi / no) + y[j]] for i in range(no)])
            for i, line in enumerate(lines):
                if with_magnitude:
                    color = hsv_to_rgb([0.0, 1, mag[int(x[j]), int(y[j])]])
                else:
                    color = hsv_to_rgb([0.0, 1, 1])
                plt.plot([y[j], line[1]], [x[j], line[0]], color=color, lw=1.0)
        plt.imshow(patch, 'gray')
    
def clean_blobs(blobs, thresh_noise=200, thresh_holes=100):
    # noise elimination
    L = label(blobs)
    regions = regionprops(L)
    for region in regions:
        a = region.area
        if a < thresh_noise:
            blobs[region.coords[:, 0], region.coords[:, 1]] = 0

    # hole filling
    # contours of blobs
    contours = ndimage.distance_transform_edt(blobs)
    contours[contours != 1] = 0
    # Small islands elimination (i.e. holes) - using a hole filling function leads to non relevant results (ex : 'D' or 'O' shapes)
    cont_l = label(1 - contours, connectivity=1)  # labelled regions without contours : let appear small island (conn 4)
    regions = regionprops(cont_l)
    for region in regions:
        if region.area < thresh_holes:  # what threshold ?
            blobs[region.coords[:, 0], region.coords[:, 1]] = 1  # keep small regions
    return blobs

def blobs_thickness(blobmap):
    distances = ndimage.distance_transform_edt(blobmap)
    skel = skeletonize(blobmap) * distances
    lambdas = 2 * skel[np.nonzero(skel)]
    return list(lambdas)